//==========================================================================
//
//      bcm53xx_serial.c
//
//      Serial device driver for Broadcom 53xx reference platform on-chip serial devices
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    msieweke, based on mips idt driver by tmichals, based on driver by dmoseley, based on POWERPC driver by jskov
// Contributors: gthomas, jskov, dmoseley, tmichals
// Date:         2003-07-02
// Purpose:      Broadcom 53xx reference platform serial device driver
// Description:  Broadcom 53xx serial device driver
//
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/io_serial.h>
#include <pkgconf/io.h>

#include <cyg/io/io.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/io/devtab.h>
#include <cyg/infra/diag.h>
#include <cyg/io/serial.h>

#ifdef CYGPKG_IO_SERIAL_MIPS_BCM53XX

#include "bcm53xx_serial.h"

cyg_uint8 gSerialPortEnabled __attribute__ ((weak)) = 1;

extern cyg_uint32 *ChipCcore;

typedef struct bcm53xx_serial_info
{
    CYG_ADDRWORD   base;
    CYG_WORD       interruptNumber;
} bcm53xx_serial_info;

static bool          bcm53xx_serial_init(struct cyg_devtab_entry *tab);
static bool          bcm53xx_serial_putc(serial_channel *chan, unsigned char c);
static Cyg_ErrNo     bcm53xx_serial_lookup(struct cyg_devtab_entry **tab, 
                                           struct cyg_devtab_entry *sub_tab,
                                           const char *name);
static unsigned char bcm53xx_serial_getc(serial_channel *chan);
static Cyg_ErrNo     bcm53xx_serial_set_config(serial_channel *chan, cyg_uint32 key,
                                               const void *xbuf, cyg_uint32 *len);
static void          bcm53xx_serial_start_xmit(serial_channel *chan);
static void          bcm53xx_serial_stop_xmit(serial_channel *chan);

static void          bcm53xx_serial_ISR(void * data);

static SERIAL_FUNS(bcm53xx_serial_funs, 
                   bcm53xx_serial_putc, 
                   bcm53xx_serial_getc,
                   bcm53xx_serial_set_config,
                   bcm53xx_serial_start_xmit,
                   bcm53xx_serial_stop_xmit
                  );

static bcm53xx_serial_info bcm53xx_serial_info0 = { BCM_SER_53XX_BASE,  
                                                    UART0_SUB_IRQ};

#if CYGNUM_IO_SERIAL_MIPS_BCM53XX_SERIAL_A_BUFSIZE > 0
static unsigned char bcm53xx_serial_out_buf0[CYGNUM_IO_SERIAL_MIPS_BCM53XX_SERIAL_A_BUFSIZE];
static unsigned char bcm53xx_serial_in_buf0[CYGNUM_IO_SERIAL_MIPS_BCM53XX_SERIAL_A_BUFSIZE];

static SERIAL_CHANNEL_USING_INTERRUPTS(bcm53xx_serial_channel0,
                                       bcm53xx_serial_funs, 
                                       bcm53xx_serial_info0,
                                       CYG_SERIAL_BAUD_RATE(CYGNUM_IO_SERIAL_MIPS_BCM53XX_SERIAL_A_BAUD),
                                       CYG_SERIAL_STOP_DEFAULT,
                                       CYG_SERIAL_PARITY_DEFAULT,
                                       CYG_SERIAL_WORD_LENGTH_DEFAULT,
                                       CYG_SERIAL_FLAGS_DEFAULT,
                                       &bcm53xx_serial_out_buf0[0], 
                                       sizeof(bcm53xx_serial_out_buf0),
                                       &bcm53xx_serial_in_buf0[0], 
                                       sizeof(bcm53xx_serial_in_buf0)
                                      );
#else
static SERIAL_CHANNEL(bcm53xx_serial_channel0,
                      bcm53xx_serial_funs, 
                      bcm53xx_serial_info0,
                      CYG_SERIAL_BAUD_RATE(CYGNUM_IO_SERIAL_MIPS_BCM53XX_SERIAL_A_BAUD),
                      CYG_SERIAL_STOP_DEFAULT,
                      CYG_SERIAL_PARITY_DEFAULT,
                      CYG_SERIAL_WORD_LENGTH_DEFAULT,
                      CYG_SERIAL_FLAGS_DEFAULT
                     );
#endif

DEVTAB_ENTRY(bcm53xx_serial_io0, 
             CYGDAT_IO_SERIAL_MIPS_BCM53XX_SERIAL_A_NAME,
             0,                 // Does not depend on a lower level interface
             &cyg_io_serial_devio, 
             bcm53xx_serial_init, 
             bcm53xx_serial_lookup,     // Serial driver may need initializing
             &bcm53xx_serial_channel0
            );


extern cyg_uint32 PeripheralClockFrequency;
void WatchDogKernelShow(cyg_uint32 taskId) __attribute__ ((weak));

void WatchDogKernelShow(cyg_uint32 taskId)
{
}

// The 16550 has a serious flaw for transmit performance.  There's no way to
// know how many characters are in the FIFO.  You can only find when the FIFO
// isn't empty.  This makes it impossible to easily fill the fifo unlexx you
// keep track of how many characters you've written.  This variable tracks how
// much free space we think there is in the FIFO.
static int gUartXmitBufferAvailable = 0;

// Internal function to actually configure the hardware to desired baud rate, etc.
static bool
bcm53xx_serial_config_port(serial_channel    *chan, 
                           cyg_serial_info_t *new_config, 
                           bool               initialize)
{
    bcm53xx_serial_info *bcm53xx_chan    = (bcm53xx_serial_info *) chan->dev_priv;
    volatile char       *UART            = (char *)                bcm53xx_chan->base;
    int                  baud_rate;
    cyg_uint32           baud_divisor;
    cyg_uint8            interrupt_mask;
    cyg_uint32           UartClockDivider;

#ifdef CYGDBG_IO_INIT
    diag_printf("BCM 53XX SERIAL config\n");
#endif

    // Translate baud rate from an enum to an integer.
    baud_rate = select_baud[new_config->baud];
    if ( baud_rate == 0 )
        return false;    // Invalid baud rate selected

    // Read the UART clock divider value from the ChipCommon core.  The UART
    // clock is based on the backplane clock divided by a constant.
    UartClockDivider = ChipCcore[0xa4/4] & 0xff;
    if (( ChipCcore[0x8/4] & 1 ) == 1 )
    {
        UartClockDivider = 5;
    }

    // Calculate baudword by dividing UART clock by (16 * baud rate) and rounding up.
    baud_divisor = ((100000000 / UartClockDivider / baud_rate) + 8) / 16;

    // Disable UART interrupts while changing hardware
    interrupt_mask = UART[IER];
    UART[IER] = 0;

    // Disable TX and RX while changing baudword.
//    UART[control &= ~(BRGEN | TXEN | RXEN);

    // Set databits, stopbits and parity.
    UART[LCR]  = select_word_length[new_config->word_length] |
                 select_stop_bits  [new_config->stop] |
                 select_parity     [new_config->parity];

    // Set baud rate.
    UART[LCR] |= LCR_DLAB;      // Turn on Divisor Latch Access Bit
    UART[DLL] = baud_divisor;
    UART[DLM] = 0;
    UART[LCR] &= ~LCR_DLAB;     // Turn off Divisor Latch Access Bit

    if ( initialize )
    {
        // Reset and clear FIFOs.  Set RX fifo threshold to 14.
        UART[FCR] = 0xc7;

        // Enable interrupts only if we're operating in buffered mode.
        // Only receive interrupts are enabled here.  Transmit interrupts
        // are enabled when the "start_xmit" function is called.
        if ( chan->out_cbuf.len != 0 )
        {
            UART[IER] = 
                IER_RCV    |       // RX data available
                IER_LS;            // RX line status (usually set by CTRL-break)
            UART[MCR] |= MCR_INT;  // Enable interrupts.
            // Enable general UART interrupt.
            hal_enable_interrupt( bcm53xx_chan->interruptNumber );
        }
        else
        {
            UART[IER] = 0;
        }
    }
    else
    {
        UART[IER] = interrupt_mask;
    }

    if ( new_config != &chan->config )
    {
        chan->config = *new_config;
    }
    return true;
}

// Function to initialize the device.  Called at bootstrap time.
static bool 
bcm53xx_serial_init(struct cyg_devtab_entry *tab)
{
    serial_channel      *chan         = (serial_channel *)      tab->priv;
    bcm53xx_serial_info *bcm53xx_chan = (bcm53xx_serial_info *) chan->dev_priv;

#ifdef CYGDBG_IO_INIT
    diag_printf("BCM 53XX SERIAL init - dev: %x.%d\n", bcm53xx_chan->base, bcm53xx_chan->interruptNumber);
#endif

    // Initialize higher level serial driver.  Really only required for interrupt driven devices.
    (chan->callbacks->serial_init)(chan);

    // If we're operating in buffered mode - implies interrupt-driven config.
    if ( chan->out_cbuf.len != 0 )
    {
        // We're not using the normal interrupt model.  All internal interrupts
        // pass through ISR 0, which then calls our sub-ISR.
        hal_map_interrupt(    bcm53xx_chan->interruptNumber, bcm53xx_serial_ISR, (void *) chan );
        hal_enable_interrupt( bcm53xx_chan->interruptNumber );
    }
    // Configure port and enable UART interrupts.
    bcm53xx_serial_config_port(chan, &chan->config, true);
    return true;
}

// This routine is called when the device is "looked" up (i.e. attached)
static Cyg_ErrNo 
bcm53xx_serial_lookup(struct cyg_devtab_entry **tab, 
                      struct cyg_devtab_entry  *sub_tab,
                      const char               *name)
{
    serial_channel  *chan = (serial_channel *)(*tab)->priv;

    // Initialize higher level serial driver.  Really only required for interrupt driven devices.
    (chan->callbacks->serial_init)(chan);
    return ENOERR;
}

// Send a character to the device output buffer.
// Return 'true' if character is sent to device
static bool
bcm53xx_serial_putc(serial_channel *chan, 
                    unsigned char   c)
{
    bcm53xx_serial_info *bcm53xx_chan = (bcm53xx_serial_info *) chan->dev_priv;
    volatile char       *UART         = (char *)                bcm53xx_chan->base;
    bool                 OkToSend     = false;

    // THRE bit (usually) indicates TX FIFO empty, so it's OK to send a character.
    if (( UART[LSR] & LSR_THRE ) != 0 )
    {
        OkToSend = true;
    }
    // If IER_TXFIFO is set, the THRE bit indicates FIFO full instead of empty.
    if ( UART[IER] & IER_TXFIFO )
    {
        OkToSend = !OkToSend;
    }

    // The THRE bit (usually) only tells us if the FIFO is empty.  We keep 
    // track of FIFO free space in the gUartXmitBufferAvailable variable. 
    // This keeps us from needing an interrupt for every character.
    if (( OkToSend ) || 
        ( gUartXmitBufferAvailable > 0 ))
    {
        if ( gSerialPortEnabled )
        {
            if (gUartXmitBufferAvailable > 0)
            {
                gUartXmitBufferAvailable -= 1;
            }
            // The transmit buffer is not full, so send a character
            UART[THR] = c;
        }
        return true;
    }
    else
    {
        // Buffer is full
        return false;
    }
}

// Fetch a character from the device input buffer, waiting if necessary
static unsigned char 
bcm53xx_serial_getc(serial_channel *chan)
{
    bcm53xx_serial_info *bcm53xx_chan = (bcm53xx_serial_info *) chan->dev_priv;
    volatile char       *UART         = (char *)                bcm53xx_chan->base;
    unsigned char        ch           = 0;

    while ( ( UART[LSR] & LSR_DR ) == 0 )
    {
        // wait
    }

    ch = UART[RBR];
    if ( gSerialPortEnabled == 0 )
    {
        ch = 0;
    }
    return ch;
}

// Set up the device characteristics; baud rate, etc.
static Cyg_ErrNo
bcm53xx_serial_set_config(serial_channel *chan, 
                          cyg_uint32      key,
                          const void     *xbuf, 
                          cyg_uint32     *len)
{
    switch ( key )
    {
        case CYG_IO_SET_CONFIG_SERIAL_INFO:
            {
                cyg_serial_info_t *config = (cyg_serial_info_t *)xbuf;
                if ( *len < sizeof(cyg_serial_info_t) )
                {
                    return -EINVAL;
                }
                *len = sizeof(cyg_serial_info_t);
                if ( bcm53xx_serial_config_port(chan, config, false) != true )
                    return -EINVAL;
            }
            break;
        default:
            return -EINVAL;
    }
    return ENOERR;
}

// Enable the transmitter on the device
static void
bcm53xx_serial_start_xmit(serial_channel *chan)
{
    bcm53xx_serial_info *bcm53xx_chan = (bcm53xx_serial_info *) chan->dev_priv;
    volatile char       *UART         = (char *)                bcm53xx_chan->base;

    // Enable general UART transmit interrupt.
    UART[IER] |= IER_XMT;

    // We should not need to call this here.  TXFIFOTHOLD Interrupts are enabled, and the ISR
    // below calls this function.  However, sometimes we get called with Master Interrupts
    // disabled, and thus the ISR never runs.  This is unfortunate because it means we
    // will be doing multiple processing steps for the same thing.
    (chan->callbacks->xmt_char)(chan);
}

// Disable the transmitter on the device
static void 
bcm53xx_serial_stop_xmit(serial_channel *chan)
{
    bcm53xx_serial_info *bcm53xx_chan = (bcm53xx_serial_info *) chan->dev_priv;
    volatile char       *UART         = (char *)                bcm53xx_chan->base;

    // Disable general UART transmit interrupt.
    UART[IER] &= ~IER_XMT;
}

// Serial I/O - low level interrupt handler (ISR)
static void 
bcm53xx_serial_ISR( void *data )
{
    serial_channel      *chan         = (serial_channel *)      data;
    bcm53xx_serial_info *bcm53xx_chan = (bcm53xx_serial_info *) chan->dev_priv;
    volatile char       *UART         = (char *)                bcm53xx_chan->base;
    cyg_uint8            rxChar       = 0;
    cyg_uint8            junk;
    cyg_uint8            IIRvalue;

    // IIR bit 0 is 0 when interrupts are pending.  Each time you read from IIR,
    // it cycles through the active interrupts.  So you must only read it once
    // for each time through the loop.
    while (( (IIRvalue = UART[IIR]) & 1 ) == 0 )
    {
        switch (( IIRvalue & IIR_IDMASK ))
        {
            // "Line Status" can mean break or a receive error.
            case IIR_LS :
                // If this was a break, tickle the watchdog.  Reading LSR also 
                // clears the error.
                if ( ( UART[LSR] & LSR_BI ) != 0 )
                {
                    WatchDogKernelShow((cyg_uint32)cyg_thread_self());
                }
                break;
            // See if there are characters to receive.
            case IIR_RX :
                while ( UART[LSR] & LSR_DR )
                {
                    rxChar = UART[RBR];
                    if ( !gSerialPortEnabled )
                    {
                        rxChar = 0;
                    }
                    (chan->callbacks->rcv_char)(chan, rxChar);
                }
                break;
            // See if we're ready to transmit.  This means the FIFO is empty
            // (FIFO mode) or the THR is empty.
            case IIR_TX :
//                hal_diag_write_char( '0' + (UART[LSR] >> 4) );
//                hal_diag_write_char( '0' + (UART[LSR] & 0xf) );
                // This is the only time we know how many characters are in the
                // FIFO.  Keep track of the free space.
                // We only assume 15 free spots, because an interrupted
                // diag_printf() could fill up the last slot.
                // NOTE: The 5352 has a much larger FIFO, but we want to stay
                // compatible with the 5365.
                gUartXmitBufferAvailable = 15;
                // Let the higher-level driver send some characters.
                (chan->callbacks->xmt_char)(chan);
#if 0
                // If we're still ready to transmit, there must not be anything wating.
                // This test looks at TEMT 
                if ( UART[LSR] & LSR_TEMT )
                {
                    // Turn off transmit interrupt if nothing's waiting.
                    UART[IER] &= ~IER_XMT;
                }
#endif
                break;
            // Test for modem status error.
            case IIR_MS :
                // Read the modem status register to clear the error.
                junk = UART[MSR];
                break;
//            default :
//                diag_printf("D%2x", UART[IIR]);
        }
    }

    // The interrupt was disabled by the int0 ISR.  Need to re-enable.
    hal_enable_interrupt( bcm53xx_chan->interruptNumber );
}

#endif

//-------------------------------------------------------------------------
// EOF bcm53xx_serial.c
