//==========================================================================
//
//      hal/mips/bcm53xx/includebcm53xx_serial.h
//
//      Broadcom 53xx register definitions.
//
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    msieweke
// Contributors: 
// Date:         2003-08-08
// Purpose:      Broadcom 53xx register definitions.
// Description:  Broadcom 53xx register definitions.
//####DESCRIPTIONEND####
//==========================================================================

#ifndef BCM53XX_REGS_H
#define BCM53XX_REGS_H

// ----------------------------------------------------------------------------
// Broadcom 53xx UART registers.
// ----------------------------------------------------------------------------

// 16550 register definitions.  Since the 5365 operates in big-endian mode, and
// all registers are defined in little-endian space, the offsets must be XORed
// with 3 to get the true address.
#define UARTENDIAN(X) X^3
#define DLL         UARTENDIAN(0)             // Divisor Latch LSB
#define DLM         UARTENDIAN(1)             // Divisor Latch MSB
#define THR         UARTENDIAN(0)             // Transmitter Holding Register
#define RBR         UARTENDIAN(0)             // Receive Buffer Register
#define IER         UARTENDIAN(1)             // Interrupt Enable Register
#define FCR         UARTENDIAN(2)             // FIFO Control Register
#define IIR         UARTENDIAN(2)             // Interrupt Identity Register
#define LCR         UARTENDIAN(3)             // Line Control Register
#define MCR         UARTENDIAN(4)             // Modem Control Register
#define LSR         UARTENDIAN(5)             // Line Status Register
#define MSR         UARTENDIAN(6)             // Modem Status Register
#define SCR         UARTENDIAN(7)             // Scratch Register

// Bit definitions for Line Status Register
#define LSR_DR      0x01            // data ready
#define LSR_OE      0x02            // overrun error
#define LSR_PE      0x04            // parity error
#define LSR_FE      0x08            // framing error
#define LSR_BI      0x10            // break interrupt
#define LSR_THRE    0x20            // transmitter holding register empty
#define LSR_TEMT    0x40            // transmitter register empty
#define LSR_ERR     0x80            // any error condition

// Bit definitions for Line Control Register
#define LCR_5BIT    0x00            // 5-bit character mode
#define LCR_6BIT    0x01            // 6-bit character mode
#define LCR_7BIT    0x02            // 7-bit character mode
#define LCR_8BIT    0x03            // 8-bit character mode
#define LCR_STB     0x04            // stop bits (2 if set, 1 if cleared)
#define LCR_PEN     0x08            // parity enable
#define LCR_EPS     0x10            // even parity select
#define LCR_SP      0x20            // stick parity
#define LCR_SB      0x40            // set break
#define LCR_DLAB    0x80            // divisor latch access bit

// Bit definitions for Interrupt Enable Register
#define IER_RCV     0x01            // receive fifo below threshold
#define IER_XMT     0x02            // transmit holding register empty
#define IER_LS      0x04            // line status change
#define IER_MS      0x08            // modem status change
#define IER_TXFIFO  0x80            // thransmit fifo below threshold

// Bit definitions for Interrupt Identification Register
#define IIR_IDMASK  0x07            // interrupt ID mask
#define IIR_NONE    0x01            // no active interrupts
#define IIR_LS      0x06            // line status (receive error or break)
#define IIR_RX      0x04            // receive data available
#define IIR_TX      0x02            // transmit holding register empty
#define IIR_MS      0x00            // modem status (CTS/DSR/etc.)

// Bit definitions for Modem Status Register
#define MSR_DCTS    0x01            // delta clear to send
#define MSR_DDSR    0x02            // delta data set ready
#define MSR_TERI    0x04            // trailing edge ring indicator
#define MSR_DDCD    0x08            // delta data carrier detect
#define MSR_CTS     0x10            // clear to send
#define MSR_DSR     0x20            // data set ready
#define MSR_RI      0x40            // ring indicator
#define MSR_DCD     0x80            // data carrier detect

// Bit definitions for Modem Control Register
#define MCR_DTR     0x01
#define MCR_RTS     0x02
#define MCR_INT     0x08            // Enable interrupts

// ----------------------------------------------------------------------------
// SDRAM Controller registers.
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// SB Core registers.
// ----------------------------------------------------------------------------

// It's easiest to think in byte addresses, but these are used as
// offsets into a uint32 array.  So divide by 4 to get array element.
#define SB_CoreControl      (0x00 >> 2)
#define SB_Config           (0x04 >> 2)
#define SB_IntStatus        (0x20 >> 2)
#define SB_IntEnableMask    (0x24 >> 2)
#define SB_IPSFlag          (0xf08 >> 2)
#define SB_IPSFlagHigh      (0xf0c >> 2)
#define SB_TPSFlag          (0xf18 >> 2)
#define SB_IntVec           (0xf94 >> 2)
#define SB_TMStateLow       (0xf98 >> 2)
#define SB_TMStateHigh      (0xf9c >> 2)
#define SB_BConfig          (0xfc0 >> 2)
#define SB_BState           (0xfc8 >> 2)
#define SB_IDLow            (0xff8 >> 2)
#define SB_IDHigh           (0xffc >> 2)

#define MIPS_Timer0         (0x28 >> 2)


// ----------------------------------------------------------------------------
// Interrupt numbers.
// ----------------------------------------------------------------------------

// The UART is treated specially.  Its interrupt comes through core 0, but
// there's no way to tell if it is active.
#define UART0_SUB_IRQ 7

// ----------------------------------------------------------------------------
// Define locations for registers
// ----------------------------------------------------------------------------

// Some BCM53xx chips have dual UARTs, but we're only using one.
#define BCM_SER_53XX_BASE       0xb8000300  // UART0 registers
#define BCM_SER_53XX_BASE1      0xb8000320  // UART1 registers

#endif // !defined( BCM53XX_REGS_H )
